"""Package for p2.loader."""
import os
import json
import logging
from datetime import datetime
from abc import ABC, abstractmethod

from p2.store.client import Client
from p2.logging import HttpLogger
from p2.tools.json import serializable

ORG='org'
REGION='region'
CLUSTER='cluster'

class Resource:

    def __init__(self,
                 layer:str,
                 kind:str,
                 name:str,
                 tag:str=None,
                 profile:dict=None,
                 children:list=None):
        self.layer = layer
        self.kind = kind
        self.name = name
        self.tag = tag
        self.profile = profile
        self.children = children or []


class Loader(ABC):
    """ Abstract base loader. Defines attributes and interfaces """

    def __init__(self, root: Resource=None, layer:str=None, tag:str=None):

        self.root  = root
        self.layer = layer
        self.tag   = tag

        p2_store_url = os.getenv('P2_STORE_URL')
        if p2_store_url is None:
            raise Exception("P2 store URL not defined")

        self.client = Client(p2_store_url)
        if self.client is None:
            raise Exception("Not able to get a p2 store client object")

        log_obj = HttpLogger(
            level = logging.DEBUG,
            silent = False,
            name='p2.loader'
        )
        self.log = log_obj.getLogger()


    @abstractmethod
    def get_resources(self):
        """
        Implemented by specific drivers.
        A driver should valorize the root object attribute.
        Clients should trap exceptions.
        """
        pass

    def sync_resources(self):
        """ Creates/updates resources into p2.store. """
        self.store_resource(self.root)
        self._delete_dead_ones(self.root)
        return 1

    def _delete_dead_ones(self, root: Resource):
        """ reorganise resources in dict with key kind, tag, name.
        For each resource kind/tag:
        - gets list of resources with same (kind, tag) from p2.store
        - for each p2.store resource found:
            if {kind, tag, name} not exists => delete record
        """
        self.res_as_list = list()
        self._traverse_tree(root)

        root_as_dict = serializable(root)
        filter = dict()
        filter['layer'] = self.layer
        filter['status'] = 'active'
        if self.tag:
            filter['tag'] = self.tag
        p2store_res = self.client.select(
            table='resources',
            filter=filter,
            fetch=['id', 'layer','kind','tag','name']
        )
        if p2store_res is None:
            raise Exception("Error getting resources")

        for res in p2store_res:
            match=dict()
            for k in ('layer', 'kind', 'tag', 'name'):
                match[k] = res[k]
            if match not in self.res_as_list:
                self.log.info(f"Going to delete resource: {res['id']}")
                ret = self.client.update(
                    table='resources',
                    record={
                        'status':'deleted',
                        'last_updated':datetime.now().isoformat()
                    },
                    where={ 'id':res['id'] }
                )
                if ret is None:
                    self.log.error(f"Could not update status of resource {res['id']}")


    def store_resource(self, resource: Resource=None, parent_id:int=None) -> int:
        """Stores the resource along with all its children
        returns the stored resource id.
        """

        res_as_dict = serializable(resource or self.root)
        res_as_dict['parent_id'] = parent_id
        res_as_dict['last_updated'] = datetime.now().isoformat()
        res_as_dict={k:v for k, v in res_as_dict.items() if v is not None}
        if 'profile' in res_as_dict:
            res_as_dict['profile'] = json.dumps(res_as_dict['profile'])

        res_children = res_as_dict.pop('children', None)

        key_cols=list(set(list(res_as_dict.keys())).intersection([
            'layer','kind','name', 'tag'
        ]))

        res_id = None
        resources = self.client.select(
            table='resources',
            fetch=['id'],
            filter={k:res_as_dict[k] for k in key_cols}
        )

        if resources is None or len(resources) == 0:
            res_id = self.client.add_returning_key(
                table='resources',
                key='id',
                record=res_as_dict)
            if res_id is None:
                self.log.error(f"Error adding resource {res_as_dict['name']}")
                raise Exception(f"Error adding resource {res_as_dict['name']}")
            else:
                self.log.info(f"Resource {res_as_dict['name']} added")
        else:
            res_id = resources[0]['id']
            ret = self.client.update(
                table='resources',
                record=res_as_dict,
                where={'id': res_id}
            )
            if (ret != 1):
                self.log.error(f"Error updating resource {res_as_dict['name']}")
                raise Exception(f"Error updating resource {res_as_dict['name']}")
            else:
                self.log.info(f"Resource {res_as_dict['name']} updated")

        if resource.children:
            for child in resource.children:
                self.store_resource(resource=child, parent_id=res_id)


    def _traverse_tree(self, res: Resource):
        self.res_as_list.append({
            'layer': res.layer,
            'kind':res.kind,
            'tag': res.tag,
            'name': res.name
        })
        for child in res.children:
            self._traverse_tree(child)
